/*
 * Dombey Copyright (c) 2010, James Bailie.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Multi-process SCGI application server as a library.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/event.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/wait.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>

#include <pwd.h>
#include <grp.h>
#include <signal.h>

#include <syslog.h>
#include <stdarg.h>
#include <string.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#define VERSION "1.5"

#define NOT_BUSY 0
#define BUSY 1
#define EMPTY -1

#define MAX_ENV 50
#define MAX_PARAM 50
#define MAX_COOKIE 50

extern char *optarg;

int scgi_fd = -1, scgi_testing = 0, scgi_backlog = 1000, scgi_min_slaves = 3,
   scgi_max_body = 10000, scgi_logging = 0, scgi_max_slaves = 25, scgi_chroot = 0,
   scgi_slave_no = -1, scgi_header_len = 0, scgi_form_data = 0;

volatile int scgi_killed = 0;

pid_t *scgi_slaves = NULL;

char *scgi_interface = "", *scgi_port = "4000", *scgi_grp = "nobody",
   *scgi_user = "nobody", *scgi_busymap = NULL, *scgi_header = NULL,
   *scgi_body = NULL, *scgi_root_dir = NULL,

   *scgi_env[ MAX_ENV + 1 ], *scgi_params[ MAX_PARAM + 1 ],
   *scgi_cookies[ MAX_COOKIE + 1 ],

   *scgi_config_file = NULL,
   *scgi_query_string = NULL, *scgi_http_cookie = NULL,
   *scgi_app_name = "LibDombey", *scgi_pidfile = "LibDombey.pid";

struct passwd *scgi_passwd;
struct group *scgi_group;

char *scgi_get_env( char * );
char **scgi_get_envs();

char *scgi_get_param( char * );
char **scgi_get_params();

char *scgi_get_cookie( char * );
char **scgi_get_cookies();

void scgi_set_name( char * );
void scgi_init_func();
void scgi_child_init_func();

void scgi_request_handler();
char *scgi_form_encode( char * );

int scgi_stack_inc = 128;

struct scgi_string
{
   int free, used;
   char *top;
   char *str;
};

void scgi_string_free( struct scgi_string *s )
{
   free( s->str );
   free( s );
}

void *scgi_memory( int size )
{
   void *ptr;

   if ( size == 0 )
      return NULL;

   if (( ptr = malloc( size )) == NULL )
   {
      if ( scgi_logging )
         syslog( LOG_WARNING, "malloc(): %m" );
      else
         fprintf( stderr, "dombey: malloc(): %s\n", strerror( errno ));

      exit( 1 );
   }

   return ptr;
}

char *scgi_str_dup( char *str, int len )
{
   char *ptr;

   if ( len < 0 )
      len = strlen( str );

   ptr = ( char *)scgi_memory( len + 1 );
   bcopy( str, ptr, len );
   ptr[ len ] = '\0';

   return ptr;
}

struct scgi_string *scgi_make_string()
{
   struct scgi_string *s;

   s = ( struct scgi_string *)scgi_memory( sizeof( struct scgi_string ));
   s->str = ( char *)scgi_memory( scgi_stack_inc + 1 );
   *s->str = '\0';
   s->free = scgi_stack_inc;
   s->used = 0;
   s->top = s->str;

   return s;
}

void scgi_string_append( struct scgi_string *s, char c )
{
   if ( s->free == 0 )
   {
      s->str = ( char *)realloc( s->str, s->used + 1 + scgi_stack_inc );

      if ( s->str == NULL )
      {
         if ( scgi_logging )
            syslog( LOG_WARNING, "realloc(): %m" );
         else
            fprintf( stderr, "dombey: realloc(): %s\n", strerror( errno ));
      }

      /* Leave room for end-of-string sentinel */

      s->free = scgi_stack_inc;
      s->top = &s->str[ s->used ];
   }

   ++s->used;
   --s->free;
   *s->top++ = c;
   *s->top = '\0';
}

void scgi_string_truncate( struct scgi_string *s )
{
   if ( s->used )
   {
      s->free += s->used;
      s->used = 0;
      *s->str = '\0';
      s->top = s->str;
   }
}

char *scgi_form_decode( char *str )
{
   static struct scgi_string *s = NULL;
   char *ptr;

   if ( s == NULL )
      s = scgi_make_string();
   else
      scgi_string_truncate( s );

   for( ptr = str; *ptr; ++ptr )
      switch( *ptr )
      {
         case '+':
            scgi_string_append( s, ' ' );
            break;

         case '%':
            if ( ! *( ptr + 1 ) || ! *( ptr + 2 ))
               scgi_string_append( s, *ptr );
            else
            {
               char data[ 3 ];
               char code;

               data[ 0 ] = *( ptr + 1 );
               data[ 1 ] = *( ptr + 2 );
               data[ 2 ] = '\0';

               if (( code = ( char )strtol( data, NULL, 16 )))
               {
                  scgi_string_append( s, code );
                  ptr += 2;
               }
               else
                  scgi_string_append( s, *ptr );
            }
            break;

         default:
            scgi_string_append( s, *ptr );
      }

   return scgi_str_dup( s->str, s->used );
}

char *scgi_form_encode( char *str )
{
   char *ptr, *ptr2;
   static struct scgi_string *s = NULL;
   int found;
   static char *reserved = "]$;:@&+!?=#{}/[^`~\"<>|%\\\t\r\n",
               *letters  = "0123456789ABCDEF";

   if ( s == NULL )
      s = scgi_make_string();
   else
   {
      scgi_string_truncate( s );
      scgi_string_append( s, '"' );
   }

   for( ptr = str; *ptr; ++ptr )
   {
      found = 0;

      if ( *ptr == ' ' )
      {
         scgi_string_append( s, '+' );
         continue;
      }

      for( ptr2 = reserved; *ptr2; ++ptr2 )
         if ( *ptr == *ptr2 )
         {
            scgi_string_append( s, '%' );
            scgi_string_append( s, letters[ *ptr / 16 ] );
            scgi_string_append( s, letters[ *ptr % 16 ] );
            found = 1;
            break;
         }

      if ( ! found )
         scgi_string_append( s, *ptr );
   }

   return scgi_str_dup( s->str, s->used );
}

/*
 * If we are the master process, we send our children SIGKILLS so they
 * will exit immediately.  SIGTERM is used by manage_slaves() to tell
 * slaves to exit at their next convenience.

 * If we are a slave process, we set the killed flag, and when we are
 * finished servicing any current connection, we will exit.  It would be
 * nice to unlink our pidfile here, but we no longer are root and so
 * cannot access /var/run/
 */

void scgi_signal_handler( int signo )
{
   int i;

   if ( scgi_slave_no >= 0 )
      scgi_killed = 1;
   else
   {
      for( i = 0; i < scgi_max_slaves; ++i )
         if ( scgi_slaves[ i ] )
            kill( scgi_slaves[ i ], SIGKILL );

      exit( 0 );
   }
}

/*
 * x - Do not become daemon.  Run single process iterative server.
 * c - chroot

 * r - root directory
 * m - max slaves
 * n - min free slaves
 * p - port

 * i - interface
 * u - user
 * g - group
 * b - max size of request bodies

 * f - config file
 */

void scgi_set_options( int argc, char **argv )
{
   int i;

   while(( i = getopt( argc, argv, "xcf:r:m:b:n:p:i:u:g:")) != -1 )
      switch( i )
      {
         case '?':
            exit( 1 );

         case 'b':
            scgi_max_body = ( optarg == NULL ? 10000 : strtol( optarg, NULL, 10 ));
            break;

         case 'c':
            scgi_chroot = 1;
            break;

         case 'f':
            scgi_config_file = optarg;
            break;

         case 'r':
            scgi_root_dir = optarg;
            break;

         case 'g':
            scgi_grp = ( optarg == NULL ? "nobody" : optarg );
            break;

         case 'i':
            scgi_interface = ( optarg == NULL ? "" : optarg );
            break;

         case 'm':
            scgi_max_slaves = ( optarg == NULL ? 25 : strtol( optarg, NULL, 10 ));
            break;

         case 'n':
            scgi_min_slaves = ( optarg == NULL ? 3 : strtol( optarg, NULL, 10 ));
            break;

         case 'p':
            scgi_port = ( optarg == NULL ? "4000" : optarg );
            break;

         case 'u':
            scgi_user = ( optarg == NULL ? "nobody" : optarg );
            break;

         case 'x':
            scgi_testing = 1;
      }

   if ( scgi_root_dir != NULL )
   {
      if ( chdir( scgi_root_dir ) < 0 )
      {
         fprintf( stderr, "%s: chdir( %s ): %s\n",
                  scgi_app_name, scgi_root_dir, strerror( errno ));
         exit( 1 );
      }
   }

   if ( scgi_chroot )
   {
      if ( scgi_root_dir != NULL )
      {
         if ( chroot( scgi_root_dir ))
         {
            fprintf( stderr, "%s: chroot( %s ): %s\n",
                     scgi_app_name, scgi_root_dir, strerror( errno ));
            exit( 1 );
         }
      }
      else
      {
         char buffer[ MAXPATHLEN + 1 ];

         if ( getcwd( buffer, MAXPATHLEN + 1 ) == NULL )
         {
            fprintf( stderr, "%s: getcwd(): %s\n",
                     scgi_app_name, strerror( errno ));
            exit( 1 );
         }

         if ( chroot( buffer ))
         {
            fprintf( stderr, "%s: chroot( %s ): %s\n",
                     scgi_app_name, buffer, strerror( errno ));
            exit( 1 );
         }
      }
   }

   if ( scgi_min_slaves <= 0 )
   {
      fprintf( stderr, "%s: min slaves <= 0: %d\n", scgi_app_name,
               scgi_min_slaves );
      exit( 1 );
   }

   if ( scgi_max_slaves <= 0 )
   {
      fprintf( stderr, "%s: max slaves <= 0: %d\n", scgi_app_name,
               scgi_max_slaves );
      exit( 1 );
   }

   if ( scgi_min_slaves > scgi_max_slaves )
   {
      fprintf( stderr, "%s: min slaves > max_slaves: %d > %d\n",
               scgi_app_name, scgi_min_slaves, scgi_max_slaves );
      exit( 1 );
   }

   if ( scgi_max_body <= 0 )
   {
      fprintf( stderr, "%s: max_body <= 0: %d\n", scgi_app_name,
               scgi_max_body );
      exit( 1 );
   }

   if (( scgi_passwd = getpwnam( scgi_user )) == NULL )
   {
      fprintf( stderr, "%s: user \"%s\" does not exist\n", scgi_app_name,
               scgi_user );
      exit( 1 );
   }

   if (( scgi_group = getgrnam( scgi_grp )) == NULL )
   {
      fprintf( stderr, "%s: group \"%s\" does not exist\n", scgi_app_name,
               scgi_grp );
      exit( 1 );
   }
}

void scgi_become_daemon()
{
   int file;

   /*
    * Fork and let the parent die, continuing as child so we are not
    * a process group leader.  This is necessary for the call to setsid().
    */

   switch( fork() )
   {
      case -1:
         fprintf( stderr, "%s: fork(): %s\n", scgi_app_name,
                  strerror( errno ));
         exit( 1 );

      case 0:
         break;

      default:
         exit( 0 );
   }

   fclose( stdout );
   fclose( stderr );
   fclose( stdin );

   stdin = fopen( "/dev/null", "r" );
   stdout = fopen( "/dev/null", "w" );
   stderr = fopen( "/dev/null", "w" );

   if ( stdin == NULL || stdout == NULL || stderr == NULL )
   {
      syslog( LOG_ERR, "fopen(): %m" );
      exit( 1 );
   }

   /*
    * Detach us from our controlling terminal, so job control and other
    * signals may not be sent to us from that terminal.
    */

   if ( setsid() < 0 )
   {
      syslog( LOG_ERR, "setsid(): %m" );
      exit( 1 );
   }

   /*
    * Write our pid to disk.
    */

   if ( ! scgi_testing )
   {
      if (( file = open( scgi_pidfile, O_WRONLY | O_CREAT | O_TRUNC,
                         S_IWUSR | S_IRUSR | S_IRGRP )) < 0 )
         syslog( LOG_WARNING, "open(): %m" );
      else
      {
         char buffer[ 16 ];

         snprintf( buffer, sizeof( buffer ), "%d", getpid() );
         write( file, buffer, strlen( buffer ));
         close( file );
      }
   }

   /*
    * We don't fork children which create files, but if we did, this will
    * make those files not world-readable by default.
    */

   umask( 2 );
}

void scgi_change_identity()
{
   if ( setgid( scgi_group->gr_gid ) < 0 )
      syslog( LOG_ERR, "setgid(): %m" );

   if ( setuid( scgi_passwd->pw_uid ) < 0 )
      syslog( LOG_ERR, "setuid(): %m" );
}

void scgi_close_on_exec( int fd )
{
   if ( fcntl( fd, F_SETFD, FD_CLOEXEC ) < 0 )
      syslog( LOG_ERR, "fcntl( F_SETFD, FD_CLOEXEC ): %m" );
}

void scgi_start_listening()
{
   struct addrinfo hints, *res;
   int result;

   /*
    * This is the new, protocol-independent way of setting up a listening
    * socket.
    */

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_flags = AI_PASSIVE;
   hints.ai_socktype = SOCK_STREAM;

   /*
    * If the user has not specified an interface, we listen on the
    * IPv6 wildcard address.
    */

   if (( result = getaddrinfo( ( *scgi_interface ? scgi_interface : NULL ),
                               scgi_port, &hints, &res )))
   {
      syslog( LOG_ERR, "getaddrinfo(): %s", gai_strerror( result ));
      exit( 1 );
   }

   if ( res == NULL )
   {
      syslog( LOG_ERR, "getaddrinfo(): no interface found" );
      exit( 1 );
   }

   scgi_fd = socket( res->ai_family, res->ai_socktype, res->ai_protocol );

   if ( scgi_fd == -1 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   result = 1;

   /*
    * Allow duplicate bindings, so we can have more than one
    * server process, if necessary.
    */

   if ( setsockopt( scgi_fd, SOL_SOCKET, SO_REUSEADDR, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_REUSEADDR ): %m" );

   /*
    * Try and detect connections which go idle for long periods of time.
    */

   if ( setsockopt( scgi_fd, SOL_SOCKET, SO_KEEPALIVE, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_KEEPALIVE ): %m" );

   /*
    * Make sure we can accept IPv4 traffic as well as IPv6, if we
    * are bound to the IPv6 wildcard address.  We are turning off
    * the option because result = 0.
    */

   result = 0;

   if ( ! *scgi_interface &&
        setsockopt( scgi_fd, IPPROTO_IPV6, IPV6_BINDV6ONLY, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( IPV6_BINDV6ONLY ): %m" );

   if ( bind( scgi_fd, res->ai_addr, res->ai_addrlen ) < 0 )
   {
      syslog( LOG_ERR, "bind(): %m" );
      exit( 1 );
   }

   if ( res != NULL )
      freeaddrinfo( res );

   if ( listen( scgi_fd, scgi_backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      exit( 1 );
   }

   scgi_close_on_exec( scgi_fd );
}

/*
 * If we are a slave, and we receive a SIGTERM, we will exit at our next
 * convenience, either the next time accept_connection() is called, after
 * we have closed the previous connection, but before we accept another, or
 * if we are already blocked in accept(), immediately, when accept() is
 * interrupted by the signal.
 */

void scgi_accept_connection()
{
   struct sockaddr_storage addr;
   socklen_t len;
   int conn;

   fclose( stdin );
   fclose( stdout );

AGAIN:
   if ( scgi_killed )
      exit( 0 );

   len = sizeof( struct sockaddr_storage );
   conn = accept( scgi_fd, ( struct sockaddr *)&addr, &len );

   if ( conn < 0 )
   {
      if ( errno == EINTR )
         goto AGAIN;

      syslog( LOG_ERR, "accept(): %m" );
      exit( 1 );
   }

   scgi_close_on_exec( conn );

   if (( stdin = fdopen( conn, "r" )) == NULL )
   {
      syslog( LOG_ERR, "fdopen(): %m" );
      exit( 1 );
   }

   if (( conn = dup( conn )) < 0 )
   {
      syslog( LOG_ERR, "dup(): %m" );
      exit( 1 );
   }

   if (( stdout = fdopen( conn, "w" )) == NULL )
   {
      syslog( LOG_ERR, "fdopen(): %m" );
      exit( 1 );
   }
}

char *scgi_find_next( char *ptr )
{
   while( --scgi_header_len )
      if ( ! *ptr++ )
         break;

   if ( ! scgi_header_len )
      return NULL;

   return ptr;
}

int scgi_init_env( int len )
{
   char *ptr;
   int n;

   n = 0;

   for( ptr = scgi_header; ptr != NULL; ptr = scgi_find_next( ptr ))
   {
      scgi_env[ n++ ] = ptr;

      if ( n == MAX_ENV )
         break;
   }

   if ( n % 2 )
      return 1;

   scgi_env[ n ] = NULL;

   return 0;
}

char *scgi_get_env( char *key )
{
   char **ptr, *ptr1, *ptr2;

   for( ptr = scgi_env; *ptr != NULL; ptr += 2 )
   {
      ptr1 = *ptr;
      ptr2 = key;

      while( *ptr1 && *ptr2 && *ptr1++ == *ptr2++ )
         ;

      if ( ! *ptr1 && ! *ptr2 )
         return *++ptr;
   }

   return NULL;
}

char **scgi_get_envs()
{
   return scgi_env;
}

int scgi_read_header()
{
   char *ptr, buffer[ 64 ];
   int c;

   for( ptr = buffer; ; ++ptr )
   {
      if (( c = getchar()) == EOF )
         return 1;

      if ( c == ':' )
      {
         *ptr = '\0';
         break;
      }

      *ptr = c;
   }

   if (( scgi_header_len = strtol( buffer, NULL, 10 )) <= 0 )
      return 1;

   if (( scgi_header = scgi_memory( scgi_header_len )) == NULL )
      return 1;

   if ( fread( scgi_header, 1, scgi_header_len, stdin ) < c )
   {
      free( scgi_header );
      scgi_header = NULL;
      return 1;
   }

   if ( getchar() != ',' )
   {
      free( scgi_header );
      scgi_header = NULL;
      return 1;
   }

   return scgi_init_env( scgi_header_len );
}

int scgi_read_body()
{
   char *ptr;
   int count;

   if (( ptr = scgi_get_env( "CONTENT_TYPE" )) != NULL &&
         !strcmp( ptr, "multipart/form-data" ))
   {
      free( scgi_header );
      scgi_header = NULL;
      return 1;
   }

   for( ptr = scgi_header; *ptr; ++ptr )
      ;

   if ( !( count = strtol( ++ptr, NULL, 10 )))
      return 0;

   /*
    * We drop connections where the client is attempting
    * to give us a request body greater than max_body characters
    * in length.
    */

   if ( count < 0 || count > scgi_max_body )
   {
      free( scgi_header );
      scgi_header = NULL;
      return 1;
   }

   if (( scgi_body = scgi_memory( count + 1 )) == NULL )
   {
      free( scgi_header );
      scgi_header = NULL;
      return 1;
   }

   if ( fread( scgi_body, 1, count, stdin ) < count )
   {
      free( scgi_header );
      free( scgi_body );
      scgi_header = scgi_body = NULL;
      return 1;
   }

   scgi_body[ count ] = '\0';

   return 0;
}

void scgi_process_cookies()
{
   char *cookie, *name, *ptr, **ptr2;
   int i;

   for( ptr2 = scgi_cookies, i = 0; i <= MAX_COOKIE; ++i )
      *ptr2++ = NULL;

   cookie = NULL;

   for( ptr2 = scgi_env; ptr2 != NULL; ptr2 += 2 )
      if ( ! strcmp( *ptr2, "HTTP_COOKIE" ))
      {
         cookie = *( ptr2 + 1 );
         break;
      }

   if ( cookie == NULL || ! *cookie )
      return;

   scgi_http_cookie = scgi_str_dup( cookie, -1 );
   cookie = scgi_http_cookie;

   for( i = 0, ptr = cookie; *ptr; ++ptr )
   {
      if ( *ptr == ';' || *ptr == ',' )
      {
         if ( i >= MAX_COOKIE )
            break;

         *ptr = '\0';

         for( name = cookie; *name && *name != '='; ++name )
            ;

         if ( *name )
            *name++ = '\0';

         while( *cookie == ' ' )
            ++cookie;

         scgi_cookies[ i++ ] = cookie;
         scgi_cookies[ i++ ] = name;

         cookie = ptr + 1;
      }
   }

  if ( *cookie && i < MAX_COOKIE - 1 )
  {
      for( name = cookie; *name && *name != '='; ++name )
         ;

      if ( *name )
         *name++ = '\0';

      while( *cookie == ' ' )
         ++cookie;

      scgi_cookies[ i++ ] = cookie;
      scgi_cookies[ i++ ] = name;
   }
}

void scgi_process_params()
{
   char *query, *ptr, *param, **ptr2;
   int i;

   for( ptr2 = scgi_params, i = 0; i <= MAX_PARAM; ++i )
      *ptr2++ = NULL;

   query = NULL;

   for( ptr2 = scgi_env; ptr2 != NULL; ptr2 += 2 )
      if ( ! strcmp( *ptr2, "QUERY_STRING" ))
      {
         query = *( ptr2 + 1 );
         break;
      }

   if ( query == NULL || ! *query )
   {
      if ( scgi_body == NULL || scgi_form_data )
         return;

      query = scgi_body;
   }
   else
   {
      scgi_query_string = scgi_str_dup( query, -1 );
      query = scgi_query_string;
   }

   for( i = 0, ptr = query; *ptr; ++ptr )
   {
      if ( *ptr == '&' )
      {
         if ( i >= MAX_PARAM )
            break;

         *ptr = '\0';

         for( param = query; *param && *param != '='; ++param )
            ;

         if ( *param )
            *param++ = '\0';

         scgi_params[ i++ ] = scgi_form_decode( query );
         scgi_params[ i++ ] = scgi_form_decode( param );

         query = ptr + 1;
      }
   }

  if ( *query && i < MAX_PARAM - 1 )
  {
      for( param = query; *param && *param != '='; ++param )
         ;

      if ( *param )
         *param++ = '\0';

      scgi_params[ i++ ] = scgi_form_decode( query );
      scgi_params[ i++ ] = scgi_form_decode( param );
   }

   return;
}

char *scgi_get_param( char *param )
{
   char **ptr, *ptr1, *ptr2;

   for( ptr = scgi_params; *ptr != NULL; ptr += 2 )
   {
      ptr1 = *ptr;
      ptr2 = param;

      while( *ptr1 && *ptr2 && *ptr1++ == *ptr2++ )
         ;

      if ( ! *ptr1 && ! *ptr2 )
         return *( ptr + 1 );
   }

   return NULL;
}

char **scgi_get_params()
{
   return scgi_params;
}

char *scgi_get_cookie( char *name )
{
   char **ptr, *ptr1, *ptr2;

   for( ptr = scgi_cookies; *ptr != NULL; ptr += 2 )
   {
      ptr1 = *ptr;
      ptr2 = name;

      while( *ptr1 && *ptr2 && *ptr1++ == *ptr2++ )
         ;

      if ( ! *ptr1 && ! *ptr2 )
         return *( ptr + 1 );
   }

   return NULL;
}

char **scgi_get_cookies()
{
   return scgi_cookies;
}

void scgi_set_name( char *name )
{
   static struct scgi_string *s = NULL;
   char *ptr;

   if ( name == NULL )
      return;

   if ( s == NULL )
      s = scgi_make_string();
   else
   {
      s->free += s->used;
      s->used = 0;
      s->top = s->str;
   }

   for( ptr = "/var/run/"; *ptr; ++ptr )
      scgi_string_append( s, *ptr );

   for( ptr = name; *ptr; ++ptr )
      scgi_string_append( s, *ptr );

   for( ptr = ".pid"; *ptr; ++ptr )
      scgi_string_append( s, *ptr );

   scgi_pidfile = s->str;
   scgi_app_name = name;
}

void scgi_service_connections()
{
   char **ptr;

   for( ; ; )
   {
      scgi_busymap[ scgi_slave_no ] = NOT_BUSY;
      scgi_accept_connection();
      scgi_busymap[ scgi_slave_no ] = BUSY;

      if ( scgi_read_header() )
         continue;

      if ( scgi_read_body() )
         continue;

      scgi_process_params();
      scgi_process_cookies();
      scgi_request_handler();

      for( ptr = scgi_params; *ptr != NULL; ++ptr )
      {
         free( *ptr );
         *ptr = NULL;
      }

      free( scgi_header );

      if ( scgi_query_string != NULL )
         free( scgi_query_string );

      if ( scgi_http_cookie != NULL )
         free( scgi_http_cookie );

      if ( scgi_body != NULL )
         free( scgi_body );

      scgi_header = scgi_body = scgi_query_string = scgi_http_cookie = NULL;
   }
}

void scgi_init_busymap()
{
   int i;

   if (( scgi_busymap = mmap( 0, scgi_max_slaves, PROT_WRITE | PROT_READ,
                              MAP_ANON | MAP_SHARED, -1, 0 )) == NULL )
   {
      syslog( LOG_ERR, "mmap(): %m" );
      exit( 1 );
   }

   scgi_slaves = scgi_memory( sizeof( pid_t ) * scgi_max_slaves );

   for( i = 0; i < scgi_max_slaves; ++i )
   {
      scgi_busymap[ i ] = EMPTY;
      scgi_slaves[ i ] = 0;
   }
}

void scgi_fork_slave( int i )
{
   pid_t pid;

   switch(( pid = fork()))
   {
      case -1:
         syslog( LOG_ERR, "fork(): %m" );
         exit( 1 );

      case 0:
         scgi_slave_no = i;
         scgi_child_init_func();
         scgi_service_connections();
         exit( 0 );

      default:
         scgi_slaves[ i ] = pid;
   }
}

void scgi_manage_slaves()
{
   pid_t pid;
   int i, idle, last_idle;

   for( ; ; )
   {
      sleep( 1 );
      idle = 0;
      last_idle = -1;

      for( i = 0; i < scgi_max_slaves; ++i )
         if ( scgi_busymap[ i ] == NOT_BUSY )
         {
            last_idle = i;
            ++idle;
         }

      if ( idle > scgi_min_slaves )
      {
         kill( scgi_slaves[ last_idle ], SIGTERM );
         --idle;
      }

      while(( pid = waitpid( -1, NULL, WNOHANG )) > 0 )
      {
         for( i = 0; i < scgi_max_slaves; ++i )
            if ( pid == scgi_slaves[ i ] )
            {
               scgi_slaves[ i ] = 0;
               scgi_busymap[ i ] = EMPTY;
               break;
            }
      }

      while( idle < scgi_min_slaves )
      {
         for( i = 0; i < scgi_max_slaves; ++i )
            if ( ! scgi_slaves[ i ] )
            {
               scgi_fork_slave( i );
               ++idle;
               break;
            }

         if ( i == scgi_max_slaves )
            break;
      }
   }
}

void scgi_set_signals()
{
   struct sigaction sigact;
   int *iptr;
   int sigs[] = {
      SIGPIPE, SIGHUP, SIGQUIT, SIGUSR1, SIGUSR2, SIGALRM, SIGINT, SIGTSTP, -1
   };

   /*
    * We ignore everything but SIGTERM when ! testing.  Otherwise, we
    * catch all the signals, so we may kill our children before exiting.
    * SIGTERM is set to interrupt system calls.
    */

   sigact.sa_handler = scgi_signal_handler;
   sigemptyset( &sigact.sa_mask );
   sigact.sa_flags = 0;

   if ( sigaction( SIGTERM, &sigact, NULL ) < 0 )
      fprintf( stderr, "dombey: sigaction: %s\n", strerror( errno ));

   for( iptr = sigs; *iptr > 0; ++iptr )
      signal( *iptr, ( scgi_testing ? scgi_signal_handler : SIG_IGN ));
}

void scgi_fork_slaves()
{
   int i;

   for( i = 0; i < scgi_min_slaves; ++i )
      scgi_fork_slave( i );
}

int main( int argc, char **argv )
{
   scgi_init_func();

   scgi_set_options( argc, argv );
   scgi_set_signals();

   openlog( scgi_app_name, LOG_PID, LOG_DAEMON );
   scgi_logging = 1;

   if ( !scgi_testing )
      scgi_become_daemon();

   scgi_start_listening();
   scgi_change_identity();
   scgi_init_busymap();

   if ( scgi_testing )
   {
      scgi_child_init_func();
      scgi_service_connections();
   }
   else
   {
      scgi_fork_slaves();
      scgi_manage_slaves();
   }

   return 0;
}
